@@ -22,7 +22,6 @@ from enum import Enum
 from random import Random, getrandbits
 from weakref import WeakKeyDictionary
 from functools import total_ordering
-from collections import defaultdict
 
 import attr
 
@@ -32,7 +31,6 @@ from hypothesis._settings import local_settings, note_deprecation
 from hypothesis.reporting import debug_report
 from hypothesis.internal.compat import Counter, ceil, hbytes, hrange, \
     int_to_bytes, benchmark_time, int_from_bytes, to_bytes_sequence
-from hypothesis.utils.conventions import UniqueIdentifier
 from hypothesis.internal.healthcheck import fail_health_check
 from hypothesis.internal.conjecture.data import MAX_DEPTH, Status, \
     StopTest, ConjectureData
@@ -47,6 +45,9 @@ __tracebackhide__ = True
 HUNG_TEST_TIME_LIMIT = 5 * 60
 MAX_SHRINKS = 500
 
+CACHE_RESET_FREQUENCY = 1000
+MUTATION_POOL_SIZE = 100
+
 
 @attr.s
 class HealthCheckState(object):
@@ -93,6 +94,24 @@ class ConjectureRunner(object):
 
         self.target_selector = TargetSelector(self.random)
 
+        self.interesting_examples = {}
+        self.covering_examples = {}
+
+        self.shrunk_examples = set()
+
+        self.health_check_state = None
+
+        self.used_examples_from_database = False
+        self.reset_tree_to_empty()
+
+    def reset_tree_to_empty(self):
+        # Previously-tested byte streams are recorded in a prefix tree, so that
+        # we can:
+        # - Avoid testing the same stream twice (in some cases).
+        # - Avoid testing a prefix of a past stream (in some cases),
+        #   since that should only result in overrun.
+        # - Generate stream prefixes that we haven't tried before.
+
         # Tree nodes are stored in an array to prevent heavy nesting of data
         # structures. Branches are dicts mapping bytes to child nodes (which
         # will in general only be partially populated). Leaves are
@@ -116,10 +135,10 @@ class ConjectureRunner(object):
         # point. Corresponds to data.write() calls.
         self.forced = {}
 
-        # Maps tree indices to the maximum byte that is valid at that point.
-        # Currently this is only used inside draw_bits, but it potentially
+        # Maps tree indices to a mask that restricts bytes at that point.
+        # Currently this is only updated by draw_bits, but it potentially
         # could get used elsewhere.
-        self.capped = {}
+        self.masks = {}
 
         # Where a tree node consists of the beginning of a block we track the
         # size of said block. This allows us to tell when an example is too
@@ -129,17 +148,6 @@ class ConjectureRunner(object):
         # buffer contents.
         self.block_sizes = {}
 
-        self.interesting_examples = {}
-        self.covering_examples = {}
-
-        self.shrunk_examples = set()
-
-        self.tag_intern_table = {}
-
-        self.health_check_state = None
-
-        self.used_examples_from_database = False
-
     def __tree_is_exhausted(self):
         return 0 in self.dead
 
@@ -170,45 +178,88 @@ class ConjectureRunner(object):
 
         self.debug_data(data)
 
-        tags = frozenset(data.tags)
-        data.tags = self.tag_intern_table.setdefault(tags, tags)
-
         if data.status == Status.VALID:
             self.valid_examples += 1
-            for t in data.tags:
-                existing = self.covering_examples.get(t)
-                if (
-                    existing is None or
-                    sort_key(data.buffer) < sort_key(existing.buffer)
-                ):
-                    self.covering_examples[t] = data
-                    if self.database is not None:
-                        self.database.save(self.covering_key, data.buffer)
-                        if existing is not None:
-                            self.database.delete(
-                                self.covering_key, existing.buffer)
 
+        # Record the test result in the tree, to avoid unnecessary work in
+        # the future.
+
+        # The tree has two main uses:
+
+        # 1. It is mildly useful in some cases during generation where there is
+        #    a high probability of duplication but it is possible to generate
+        #    many examples. e.g. if we had input of the form none() | text()
+        #    then we would generate duplicates 50% of the time, and would
+        #    like to avoid that and spend more time exploring the text() half
+        #    of the search space. The tree allows us to predict in advance if
+        #    the test would lead to a duplicate and avoid that.
+        # 2. When shrinking it is *extremely* useful to be able to anticipate
+        #    duplication, because we try many similar and smaller test cases,
+        #    and these will tend to have a very high duplication rate. This is
+        #    where the tree usage really shines.
+        #
+        # Unfortunately, as well as being the less useful type of tree usage,
+        # the first type is also the most expensive! Once we've entered shrink
+        # mode our time remaining is essentially bounded - we're just here
+        # until we've found the minimal example. In exploration mode, we might
+        # be early on in a very long-running processs, and keeping everything
+        # we've ever seen lying around ends up bloating our memory usage
+        # substantially by causing us to use O(max_examples) memory.
+        #
+        # As a compromise, what we do is reset the cache every so often. This
+        # keeps our memory usage bounded. It has a few unfortunate failure
+        # modes in that it means that we can't always detect when we should
+        # have stopped - if we are exploring a language which has only slightly
+        # more than cache reset frequency number of members, we will end up
+        # exploring indefinitely when we could have stopped. However, this is
+        # a fairly unusual case - thanks to exponential blow-ups in language
+        # size, most languages are either very large (possibly infinite) or
+        # very small. Nevertheless we want CACHE_RESET_FREQUENCY to be quite
+        # high to avoid this case coming up in practice.
+        if (
+            self.call_count % CACHE_RESET_FREQUENCY == 0 and
+            not self.interesting_examples
+        ):
+            self.reset_tree_to_empty()
+
+        # First, iterate through the result's buffer, to create the node that
+        # will hold this result. Also note any forced or masked bytes.
         tree_node = self.tree[0]
         indices = []
         node_index = 0
         for i, b in enumerate(data.buffer):
+            # We build a list of all the node indices visited on our path
+            # through the tree, since we'll need to refer to them later.
             indices.append(node_index)
+
+            # If this buffer position was forced or masked, then mark its
+            # corresponding node as forced/masked.
             if i in data.forced_indices:
                 self.forced[node_index] = b
             try:
-                self.capped[node_index] = data.capped_indices[i]
+                self.masks[node_index] = data.masked_indices[i]
             except KeyError:
                 pass
+
             try:
+                # Use the current byte to find the next node on our path.
                 node_index = tree_node[b]
             except KeyError:
+                # That node doesn't exist yet, so create it.
                 node_index = len(self.tree)
+                # Create a new branch node. If this should actually be a leaf
+                # node, it will be overwritten when we store the result.
                 self.tree.append({})
                 tree_node[b] = node_index
+
             tree_node = self.tree[node_index]
+
             if node_index in self.dead:
+                # This part of the tree has already been marked as dead, so
+                # there's no need to traverse any deeper.
                 break
 
+        # At each node that begins a block, record the size of that block.
         for u, v in data.blocks:
             # This can happen if we hit a dead node when walking the buffer.
             # In that case we already have this section of the tree mapped.
@@ -216,21 +267,44 @@ class ConjectureRunner(object):
                 break
             self.block_sizes[indices[u]] = v - u
 
+        # Forcibly mark all nodes beyond the zero-bound point as dead,
+        # because we don't intend to try any other values there.
         self.dead.update(indices[self.cap:])
 
+        # Now store this result in the tree (if appropriate), and check if
+        # any nodes need to be marked as dead.
         if data.status != Status.OVERRUN and node_index not in self.dead:
+            # Mark this node as dead, because it produced a result.
+            # Trying to explore suffixes of it would not be helpful.
             self.dead.add(node_index)
+            # Store the result in the tree as a leaf. This will overwrite the
+            # branch node that was created during traversal.
             self.tree[node_index] = data
 
+            # Review the traversed nodes, to see if any should be marked
+            # as dead. We check them in reverse order, because as soon as we
+            # find a live node, all nodes before it must still be live too.
             for j in reversed(indices):
+                mask = self.masks.get(j, 0xff)
+                assert _is_simple_mask(mask)
+                max_size = mask + 1
+
                 if (
-                    len(self.tree[j]) < self.capped.get(j, 255) + 1 and
+                    len(self.tree[j]) < max_size and
                     j not in self.forced
                 ):
+                    # There are still byte values to explore at this node,
+                    # so it isn't dead yet.
                     break
                 if set(self.tree[j].values()).issubset(self.dead):
+                    # Everything beyond this node is known to be dead,
+                    # and there are no more values to explore here (see above),
+                    # so this node must be dead too.
                     self.dead.add(j)
                 else:
+                    # Even though all of this node's possible values have been
+                    # tried, there are still some deeper nodes that remain
+                    # alive, so this node isn't dead yet.
                     break
 
         if data.status == Status.INTERESTING:
@@ -288,23 +362,47 @@ class ConjectureRunner(object):
         self.record_for_health_check(data)
 
     def generate_novel_prefix(self):
+        """Uses the tree to proactively generate a starting sequence of bytes
+        that we haven't explored yet for this test.
+
+        When this method is called, we assume that there must be at
+        least one novel prefix left to find. If there were not, then the
+        test run should have already stopped due to tree exhaustion.
+        """
         prefix = bytearray()
         node = 0
         while True:
             assert len(prefix) < self.cap
             assert node not in self.dead
-            upper_bound = self.capped.get(node, 255) + 1
+
+            # Figure out the range of byte values we should be trying.
+            # Normally this will be 0-255, unless the current position has a
+            # mask.
+            mask = self.masks.get(node, 0xff)
+            assert _is_simple_mask(mask)
+            upper_bound = mask + 1
+
             try:
                 c = self.forced[node]
+                # This position has a forced byte value, so trying a different
+                # value wouldn't be helpful. Just add the forced byte, and
+                # move on to the next position.
                 prefix.append(c)
                 node = self.tree[node][c]
                 continue
             except KeyError:
                 pass
+
+            # Provisionally choose the next byte value.
+            # This will change later if we find that it was a bad choice.
             c = self.random.randrange(0, upper_bound)
+
             try:
                 next_node = self.tree[node][c]
                 if next_node in self.dead:
+                    # Whoops, the byte value we chose for this position has
+                    # already been fully explored. Let's pick a new value, and
+                    # this time choose a value that's definitely still alive.
                     choices = [
                         b for b in hrange(upper_bound)
                         if self.tree[node].get(b) not in self.dead
@@ -313,9 +411,13 @@ class ConjectureRunner(object):
                     c = self.random.choice(choices)
                     node = self.tree[node][c]
                 else:
+                    # The byte value we chose is in the tree, but it still has
+                    # some unexplored descendants, so it's a valid choice.
                     node = next_node
                 prefix.append(c)
             except KeyError:
+                # The byte value we chose isn't in the tree at this position,
+                # which means we've successfully found a novel prefix.
                 prefix.append(c)
                 break
         assert node not in self.dead
@@ -732,10 +834,7 @@ class ConjectureRunner(object):
             self.test_function(last_data)
             last_data.freeze()
 
-            if len(self.covering_examples) > targets_found:
-                count = 0
-            else:
-                count += 1
+            count += 1
 
         mutations = 0
         mutator = self._new_mutator()
@@ -782,7 +881,7 @@ class ConjectureRunner(object):
                 self.test_function(data)
                 data.freeze()
             else:
-                target, origin = self.target_selector.select()
+                origin = self.target_selector.select()
                 mutations += 1
                 targets_found = len(self.covering_examples)
                 data = ConjectureData(
@@ -798,7 +897,6 @@ class ConjectureRunner(object):
                     mutations = 0
                 elif (
                     data.status < origin.status or
-                    not self.target_selector.has_tag(target, data) or
                     mutations >= 10
                 ):
                     # Cap the variations of a single example and move on to
@@ -908,10 +1006,16 @@ class ConjectureRunner(object):
         This is purely an optimisation to try to reduce the number of tests we
         run. "return True" would be a valid but inefficient implementation.
         """
+
+        # Traverse the tree, to see if we have already tried this buffer
+        # (or a prefix of it).
         node_index = 0
         n = len(buffer)
         for k, b in enumerate(buffer):
             if node_index in self.dead:
+                # This buffer (or a prefix of it) has already been tested,
+                # or has already had its descendants fully explored.
+                # Testing it again would not be helpful.
                 return False
             try:
                 # The block size at that point provides a lower bound on how
@@ -922,39 +1026,73 @@ class ConjectureRunner(object):
                     return False
             except KeyError:
                 pass
+
+            # If there's a forced value or a mask at this position, then
+            # pretend that the buffer already contains a matching value,
+            # because the test function is going to do the same.
             try:
                 b = self.forced[node_index]
             except KeyError:
                 pass
             try:
-                b = b & self.capped[node_index]
+                b = b & self.masks[node_index]
             except KeyError:
                 pass
+
             try:
                 node_index = self.tree[node_index][b]
             except KeyError:
+                # The buffer wasn't in the tree, which means we haven't tried
+                # it. That makes it a possible candidate.
                 return True
         else:
+            # We ran out of buffer before reaching a leaf or a missing node.
+            # That means the test function is going to draw beyond the end
+            # of this buffer, which makes it a bad candidate.
             return False
 
     def cached_test_function(self, buffer):
+        """Checks the tree to see if we've tested this buffer, and returns the
+        previous result if we have.
+
+        Otherwise we call through to ``test_function``, and return a
+        fresh result.
+        """
         node_index = 0
         for c in buffer:
+            # If there's a forced value or a mask at this position, then
+            # pretend that the buffer already contains a matching value,
+            # because the test function is going to do the same.
             try:
                 c = self.forced[node_index]
             except KeyError:
                 pass
             try:
-                c = c & self.capped[node_index]
+                c = c & self.masks[node_index]
             except KeyError:
                 pass
+
             try:
                 node_index = self.tree[node_index][c]
             except KeyError:
+                # The byte at this position isn't in the tree, which means
+                # we haven't tested this buffer. Break out of the tree
+                # traversal, and run the test function normally.
                 break
             node = self.tree[node_index]
             if isinstance(node, ConjectureData):
+                # This buffer (or a prefix of it) has already been tested.
+                # Return the stored result instead of trying it again.
                 return node
+        else:
+            # Falling off the end of this loop means that we're about to test
+            # a prefix of a previously-tested byte stream. The test is going
+            # to draw beyond the end of the buffer, and fail due to overrun.
+            # Currently there is no special handling for this case.
+            pass
+
+        # We didn't find a match in the tree, so we need to run the test
+        # function normally.
         result = ConjectureData.for_buffer(buffer)
         self.test_function(result)
         return result
@@ -971,6 +1109,16 @@ class ConjectureRunner(object):
         return result
 
 
+def _is_simple_mask(mask):
+    """A simple mask is ``(2 ** n - 1)`` for some ``n``, so it has the effect
+    of keeping the lowest ``n`` bits and discarding the rest.
+
+    A mask in this form can produce any integer between 0 and the mask itself
+    (inclusive), and the total number of these values is ``(mask + 1)``.
+    """
+    return (mask & (mask + 1)) == 0
+
+
 def _draw_predecessor(rnd, xs):
     r = bytearray()
     any_strict = False
@@ -1007,140 +1155,53 @@ def uniform(random, n):
     return int_to_bytes(random.getrandbits(n * 8), n)
 
 
-class SampleSet(object):
-    """Set data type with the ability to sample uniformly at random from it.
+def pop_random(random, values):
+    """Remove a random element of values, possibly changing the ordering of its
+    elements."""
 
-    The mechanism is that we store the set in two parts: A mapping of
-    values to their index in an array. Sampling uniformly at random then
-    becomes simply a matter of sampling from the array, but we can use
-    the index for efficient lookup to add and remove values.
-    """
-
-    __slots__ = ('__values', '__index')
-
-    def __init__(self):
-        self.__values = []
-        self.__index = {}
-
-    def __len__(self):
-        return len(self.__values)
-
-    def __repr__(self):
-        return 'SampleSet(%r)' % (self.__values,)
-
-    def add(self, value):
-        assert value not in self.__index
-        # Adding simply consists of adding the value to the end of the array
-        # and updating the index.
-        self.__index[value] = len(self.__values)
-        self.__values.append(value)
-
-    def remove(self, value):
-        # To remove a value we first remove it from the index. But this leaves
-        # us with the value still in the array, so we have to fix that. We
-        # can't simply remove the value from the array, as that would a) Be an
-        # O(n) operation and b) Leave the index completely wrong for every
-        # value after that index.
-        # So what we do is we take the last element of the array and place it
-        # in the position of the value we just deleted (if the value was not
-        # already the last element of the array. If it was then we don't have
-        # to do anything extra). This reorders the array, but that's OK because
-        # we don't care about its order, we just need to sample from it.
-        i = self.__index.pop(value)
-        last = self.__values.pop()
-        if i < len(self.__values):
-            self.__values[i] = last
-            self.__index[last] = i
-
-    def choice(self, random):
-        return random.choice(self.__values)
-
-
-class Negated(object):
-    __slots__ = ('tag',)
-
-    def __init__(self, tag):
-        self.tag = tag
-
-
-NEGATED_CACHE = {}  # type: dict
-
-
-def negated(tag):
-    try:
-        return NEGATED_CACHE[tag]
-    except KeyError:
-        result = Negated(tag)
-        NEGATED_CACHE[tag] = result
-        return result
-
-
-universal = UniqueIdentifier('universal')
+    # We pick the element at a random index. Rather than removing that element
+    # from the list (which would be an O(n) operation), we swap it to the end
+    # and return the last element of the list. This changes the order of
+    # the elements, but as long as these elements are only accessed through
+    # random sampling that doesn't matter.
+    i = random.randrange(0, len(values))
+    values[i], values[-1] = values[-1], values[i]
+    return values.pop()
 
 
 class TargetSelector(object):
     """Data structure for selecting targets to use for mutation.
 
-    The goal is to do a good job of exploiting novelty in examples without
-    getting too obsessed with any particular novel factor.
+    The main purpose of the TargetSelector is to maintain a pool of "reasonably
+    useful" examples, while keeping the pool of bounded size.
 
-    Roughly speaking what we want to do is give each distinct coverage target
-    equal amounts of time. However some coverage targets may be harder to fuzz
-    than others, or may only appear in a very small minority of examples, so we
-    don't want to let those dominate the testing.
+    In particular it ensures:
 
-    Targets are selected according to the following rules:
+    1. We only retain examples of the best status we've seen so far (not
+       counting INTERESTING, which is special).
+    2. We preferentially return examples we've never returned before when
+       select() is called.
+    3. The number of retained examples is never more than self.pool_size, with
+       past examples discarded automatically, preferring ones that we have
+       already explored from.
 
-    1. We ideally want valid examples as our starting point. We ignore
-       interesting examples entirely, and other than that we restrict ourselves
-       to the best example status we've seen so far. If we've only seen
-       OVERRUN examples we use those. If we've seen INVALID but not VALID
-       examples we use those. Otherwise we use VALID examples.
-    2. Among the examples we've seen with the right status, when asked to
-       select a target, we select a coverage target and return that along with
-       an example exhibiting that target uniformly at random.
-
-    Coverage target selection proceeds as follows:
-
-    1. Whenever we return an example from select, we update the usage count of
-       each of its tags.
-    2. Whenever we see an example, we add it to the list of examples for all of
-       its tags.
-    3. When selecting a tag, we select one with a minimal usage count. Among
-       those of minimal usage count we select one with the fewest examples.
-       Among those, we select one uniformly at random.
-
-    This has the following desirable properties:
-
-    1. When two coverage targets are intrinsically linked (e.g. when you have
-       multiple lines in a conditional so that either all or none of them will
-       be covered in a conditional) they are naturally deduplicated.
-    2. Popular coverage targets will largely be ignored for considering what
-       test to run - if every example exhibits a coverage target, picking an
-       example because of that target is rather pointless.
-    3. When we discover new coverage targets we immediately exploit them until
-       we get to the point where we've spent about as much time on them as the
-       existing targets.
-    4. Among the interesting deduplicated coverage targets we essentially
-       round-robin between them, but with a more consistent distribution than
-       uniformly at random, which is important particularly for short runs.
+    These invariants are fairly heavily prone to change - they're not
+    especially well validated as being optimal, and are mostly just a decent
+    compromise between diversity and keeping the pool size bounded.
     """
 
-    def __init__(self, random):
+    def __init__(self, random, pool_size=MUTATION_POOL_SIZE):
         self.random = random
         self.best_status = Status.OVERRUN
+        self.pool_size = pool_size
         self.reset()
 
+    def __len__(self):
+        return len(self.fresh_examples) + len(self.used_examples)
+
     def reset(self):
-        self.examples_by_tags = defaultdict(list)
-        self.tag_usage_counts = Counter()
-        self.tags_by_score = defaultdict(SampleSet)
-        self.scores_by_tag = {}
-        self.scores = []
-        self.mutation_counts = 0
-        self.example_counts = 0
-        self.non_universal_tags = set()
-        self.universal_tags = None
+        self.fresh_examples = []
+        self.used_examples = []
 
     def add(self, data):
         if data.status == Status.INTERESTING:
@@ -1151,82 +1212,24 @@ class TargetSelector(object):
             self.best_status = data.status
             self.reset()
 
-        if self.universal_tags is None:
-            self.universal_tags = set(data.tags)
-        else:
-            not_actually_universal = self.universal_tags - data.tags
-            for t in not_actually_universal:
-                self.universal_tags.remove(t)
-                self.non_universal_tags.add(t)
-                self.examples_by_tags[t] = list(
-                    self.examples_by_tags[universal]
-                )
-
-        new_tags = data.tags - self.non_universal_tags
-
-        for t in new_tags:
-            self.non_universal_tags.add(t)
-            self.examples_by_tags[negated(t)] = list(
-                self.examples_by_tags[universal]
-            )
-
-        self.example_counts += 1
-        for t in self.tags_for(data):
-            self.examples_by_tags[t].append(data)
-            self.rescore(t)
-
-    def has_tag(self, tag, data):
-        if tag is universal:
-            return True
-        if isinstance(tag, Negated):
-            return tag.tag not in data.tags
-        return tag in data.tags
-
-    def tags_for(self, data):
-        yield universal
-        for t in data.tags:
-            yield t
-        for t in self.non_universal_tags:
-            if t not in data.tags:
-                yield negated(t)
-
-    def rescore(self, tag):
-        new_score = (
-            self.tag_usage_counts[tag], len(self.examples_by_tags[tag]))
-        try:
-            old_score = self.scores_by_tag[tag]
-        except KeyError:
-            pass
-        else:
-            self.tags_by_score[old_score].remove(tag)
-        self.scores_by_tag[tag] = new_score
-
-        sample = self.tags_by_score[new_score]
-        if len(sample) == 0:
-            heapq.heappush(self.scores, new_score)
-        sample.add(tag)
-
-    def select_tag(self):
-        while True:
-            peek = self.scores[0]
-            sample = self.tags_by_score[peek]
-            if len(sample) == 0:
-                heapq.heappop(self.scores)
-            else:
-                return sample.choice(self.random)
-
-    def select_example_for_tag(self, t):
-        return self.random.choice(self.examples_by_tags[t])
+        # Note that technically data could be a duplicate. This rarely happens
+        # (only if we've exceeded the CACHE_RESET_FREQUENCY number of test
+        # function calls), but it's certainly possible. This could result in
+        # us having the same example multiple times, possibly spread over both
+        # lists. We could check for this, but it's not a major problem so we
+        # don't bother.
+        self.fresh_examples.append(data)
+        if len(self) > self.pool_size:
+            pop_random(self.random, self.used_examples or self.fresh_examples)
+            assert self.pool_size == len(self)
 
     def select(self):
-        t = self.select_tag()
-        self.mutation_counts += 1
-        result = self.select_example_for_tag(t)
-        assert self.has_tag(t, result)
-        for s in self.tags_for(result):
-            self.tag_usage_counts[s] += 1
-            self.rescore(s)
-        return t, result
+        if self.fresh_examples:
+            result = pop_random(self.random, self.fresh_examples)
+            self.used_examples.append(result)
+            return result
+        else:
+            return self.random.choice(self.used_examples)
 
 
 def block_program(description):
@@ -2003,7 +2006,6 @@ class Shrinker(object):
 
         self.shrink_target = new_target
         self.__shrinking_block_cache = {}
-        self.__intervals = None
 
     def try_shrinking_blocks(self, blocks, b):
         """Attempts to replace each block in the blocks list with b. Returns
